////////////////////////////////////////////////////////////////
//
// File: bmptools.cpp
// Author: Sven Goossens
// Purpose:
//		This file contains the functions that you will need to 
//		load a bitmap file into your program.
//		It also contains the wrapper functions of the DCT and
//		IDCT, needed for the frequency space conversion.
//
////////////////////////////////////////////////////////////////
#include "stdio.h"
#include "stdlib.h"

#include "bmptools.h"

//---------------------------------------------------------------------------
/* unsigned char *loadbmp(char *filename, BITMAPINFOHEADER *bmpinfo)
	
	This function will load the bmp file described in 'filename'.
	The return value of this function is a pointer to a newly allocated
	piece of memory, at which the bitmap is located.

	The output will be a an array of values, representing the Red, Green and Blue pixel
	values of the BMP. See the text in the assignment for more details.

	The second input argument of the function is a pointer to a BITMAPINFOHEADER
	struct. This struct will be filled with the header information of the BMP.
	This struct will contain all the important information about the image,
	such as the width and height.	
*/
unsigned char *loadbmp(char *filename, BITMAPINFOHEADER *bmpinfo)
{
	FILE *fileptr;
	BITMAPFILEHEADER bmpfile;
	unsigned char *bmpimage;
	int imageidx = 0, x, y, i;
	unsigned char swaprgb;

	// Check the input arguments of this function
	if(filename == NULL || bmpinfo == NULL){
		printf("The input arguments of the loadbmp function aren't correctly set. Filename = %x, bmpinfo = %x. Aborting image load. \n", (int)filename, (int)bmpinfo);
		return 0;
	}

	// Attempt to open the file. Make sure it is in the working directory of you program
	// typically the same directory as the .vcproj file of your project
	fileptr = fopen(filename, "rb");
	if (fileptr == NULL){
		printf("The file could not be opened. Are you sure you placed it in the right directory?\n");
		return 0;
	}

	// Read the bitmap header into the bmpfile struct
	// The header is read as raw bytes, which can than easily be accessed by trough the struct
	// It is as if the struct sorts the data
	fread(&bmpfile, sizeof(BITMAPFILEHEADER), 1, fileptr);

	// Check if we are dealing with a bitmap
	if (bmpfile.bfType != BITMAP_ID)
	{
		printf("Invalid file detected. This is not a BMP.\n");
		fclose(fileptr);
		return 0;
	}

	// Read the Bitmap information or DIB header
	fread(bmpinfo, sizeof(BITMAPINFOHEADER), 1, fileptr);

	// If the struct contains any of the 3 entrys described below, we aren't able to handle
	// this specific type of BMP file
	if(bmpinfo->biCompression != 0){	// compressed image, no support for that
		printf("Compressed bmp image detected. Aborting load of image.\n");
		return 0;
	}
	if(bmpinfo->biPlanes != 1){			// number of planes should be 1, abort of not
		printf("Number of image planes != 1. Aborting load of image.\n");
		return 0;
	}
	if(bmpinfo->biBitCount != 24){		// 24 bits per pixel, 8 bits per color
		printf("Number of bits per pixel != 24. Aborting load of image.\n");
		return 0;
	}
/*
	fread(&my_EJPGBlock, sizeof(EJPGBlock), 1, fp);
	// Find out how many bytes actually belonged to the block, store this in 'real_size'
	fseek(fp, sizeof(EJPGBlock)-real_size, SEEK_CUR);
	*/

	// Go to the start of the pixel data
	fseek(fileptr, bmpfile.bfOffBits, SEEK_SET);

	// Allocate memory for the BMP image
	// Make sure you free this lateron!
	bmpimage=(unsigned char*)malloc(bmpinfo->biSizeImage);
	if (!bmpimage)	// Malloc failed
	{
		fclose(fileptr);
		printf("Allocating space for the pixel data failed. Aborting load of image.\n");
		return 0;
	}

	// Read the image data in the allocated memory
	// fread return the number of bytes read, which shouldn't be equal to zero at least
	if (fread(bmpimage, 1, bmpinfo->biSizeImage, fileptr) == NULL)
	{
		fclose(fileptr);
		free(bmpimage);
		printf("Reading the pixel data failed. Aborting load of image.\n");
		return 0;
	}

	// The following section of codes swaps the Red and Blue pixel values in the image
	// This is done because a BMP image stores the pixels in the BGR format, 
	// but we would like to have them as RGB.
	// The code also removes the padding bytes that are present at the end of each image line.
	imageidx = 0;
	i=0;
	for(y=0; y<bmpinfo->biHeight; y++){
		for(x=0; x<bmpinfo->biWidth; x++){
			swaprgb = bmpimage[i];
			bmpimage[imageidx]	 = bmpimage[i+2];
			bmpimage[imageidx+1] = bmpimage[i+1];
			bmpimage[imageidx+2] = swaprgb;

			imageidx += 3;
			i+=3;
		}
		i+= bmpinfo->biWidth%4;
	}

	// If we get here, the image has been succesfully loaded
	fclose(fileptr);

	// Pointer to the bitmap data.
	// YOU are responsible for freeing it!
	return bmpimage;
}

//---------------------------------------------------------------------------
/* void DCT(PBlock* P, FBlock* F)
	The DCT function. This function serves as a wrapper for the real DCT
	function, which is stored in jfdctint.cpp. 
	This function has 2 input arguments:
	- PBlock* P: A pointer to a PBlock struct, containing the spatial pixel
				 information of the block
	- FBlock* F: A pointer to an FBlock struct, containing the frequency
				 information. The output of the DCT will be written to this
				 struct.
	 Note that the PBlock struct is overwritten!
*/
	
void DCT(PBlock* P, FBlock* F){
	int i;

	jpeg_fdct_islow((int*)(&P->linear[0]));


	for(i=0; i<64; i++ ){
		F->linear[i] = P->linear[i]>>3;
	}
}
//---------------------------------------------------------------------------
/* void IDCT(FBlock* S, PBlock* T)
	The IDCT function.
	This function has 2 input arguments:
	- FBlock* S: A pointer to a FBlock struct, containing the frequency
				 information of the block
	- PBlock* T: A pointer to an FBlock struct, containing the spatial pixel
				 information. The output of the IDCT will be written to this
				 struct.
	 Note that the FBlock struct is overwritten in the process!
*/
void IDCT(FBlock* S, PBlock* T)
{
	idctrow((int*)(S->linear));
	idctrow((int*)(S->linear+8));
	idctrow((int*)(S->linear+16));
	idctrow((int*)(S->linear+24));
	idctrow((int*)(S->linear+32));
	idctrow((int*)(S->linear+40));
	idctrow((int*)(S->linear+48));
	idctrow((int*)(S->linear+56));

	idctcol((int*)(S->linear), (unsigned int*)(T->linear));
	idctcol((int*)(S->linear+1), (unsigned int*)(T->linear+1));
	idctcol((int*)(S->linear+2), (unsigned int*)(T->linear+2));
	idctcol((int*)(S->linear+3), (unsigned int*)(T->linear+3));
	idctcol((int*)(S->linear+4), (unsigned int*)(T->linear+4));
	idctcol((int*)(S->linear+5), (unsigned int*)(T->linear+5));
	idctcol((int*)(S->linear+6), (unsigned int*)(T->linear+6));
	idctcol((int*)(S->linear+7), (unsigned int*)(T->linear+7));
}
//---------------------------------------------------------------------------